from abc import ABC, abstractmethod


class THipsterException(Exception, ABC):
    @property
    @abstractmethod
    def message():
        raise NotImplementedError()

    @abstractmethod
    def __str__(self) -> str:
        return self.message



from thipster.engine.parsed_file import Position


class LexerPosition():
    def __init__(
        self,
        fileName: str = '',
        startLine: int = 1,
        startColumn: int = 1,
    ) -> None:
        """Class to represent the state and position of the lexer

        Parameters
        ----------
        fileName : str
            Name of the current file
        startLine : int, optional
            Starting line in the current file, by default 1
        startColumn : int, optional
            Starting column in the current file, by default 1
        """
        self.__currentFile = fileName
        self.__currentLine = startLine
        self.__currentColumn = startColumn
        self.__currentToken = ''
        self.__currentTokenIndex = startColumn
        self.__currentTokenLine = startColumn
        self.__isVariable = False
        self.__isQuotedString = False
        self.__isMultiLine = False
        self.__consecutiveWhitespaces = 0

    @property
    def currentFile(self):
        return self.__currentFile

    @property
    def currentLine(self):
        return self.__currentLine

    @property
    def currentColumn(self):
        return self.__currentColumn

    @property
    def currentToken(self):
        return self.__currentToken

    @property
    def currentTokenIndex(self):
        return self.__currentTokenIndex

    @property
    def isCurrentTokenAVariable(self):
        return self.__isVariable

    @property
    def isCurrentTokenMultiLine(self):
        return self.__isMultiLine

    @property
    def isQuotedString(self):
        return self.__isQuotedString

    @isQuotedString.setter
    def isQuotedString(self, value):
        self.__isQuotedString = value

    @property
    def consecutiveWhitespaces(self):
        return self.__consecutiveWhitespaces

    @property
    def currentCharPosition(self) -> Position:
        return Position(
            self.__currentFile,
            self.__currentLine,
            self.__currentColumn,
        )

    @property
    def currentTokenPosition(self) -> Position:
        return Position(
            self.__currentFile,
            self.__currentTokenLine,
            self.__currentTokenIndex,
        )

    def newLine(self) -> None:
        """Get the position to the next line
        """
        self.__currentLine += 1
        self.__currentColumn = 1

    def nextColumn(self, step: int = 1) -> None:
        """Get the position to a new column

        Parameters
        ----------
        step : int
            Step to modify the current column, by default 1
        """
        if (self.__currentColumn + step) > 0:
            self.__currentColumn += step
        else:
            self.__currentColumn = 0  # Raise exception ?

    def addCharToCurrentToken(self, char) -> None:
        """Add a char to the current stored token

        Parameters
        ----------
        char : _type_
            Char to add to the token
        """
        self.__currentToken += char

    def setCurrentTokenIndex(self, newIndex: int | None = None) -> None:
        """Modify the stored token index

        Parameters
        ----------
        newIndex : int, optional
            New index of the stored token, by default None
        """
        if newIndex:
            self.__currentTokenIndex = newIndex
        else:
            self.__currentTokenIndex = self.__currentColumn
            self.__currentTokenLine = self.__currentLine

    def resetCurrentToken(self, newIndex: int | None = None) -> None:
        """Reset the current stored token and its index

        Parameters
        ----------
        newIndex : int, optional
            New index of the stored token, by default None
        """
        self.__currentToken = ''
        self.setCurrentTokenIndex(newIndex)

    def addConsecutiveWhitespace(self) -> None:
        """Add a consecutive whitespace to the lexer state
        """
        self.__consecutiveWhitespaces += 1

    def resetConsecutiveWhitespaces(self) -> None:
        """Reset consecutive whitespaces of the lexer state to 0
        """
        self.__consecutiveWhitespaces = 0

    def setIsVariable(self, value: bool = False) -> None:
        """Set the current stored token as a variable

        Parameters
        ----------
        value : bool, optional
            Is the stored token a variable True or False, by default False
        """
        self.__isVariable = value

    def setIsMultiLine(self, value: bool = False) -> None:
        """Set the current token as a multi-line input

        Parameters
        ----------
        value : bool, optional
            Is the current token a multi-line input True or False, by default False
        """
        self.__isMultiLine = value



import os

from thipster.engine import I_Parser
from thipster.engine.parsed_file import ParsedFile

from .exceptions import DSLParserBaseException, DSLParserPathNotFound
from .interpreter import Interpreter
from .lexer import Lexer
from .token_parser import TokenParser


class DSLParser(I_Parser):

    def __getfiles(path: str) -> dict[str, str]:
        """Recursively get all files in the requested directory and its sudirectories
        Can be run on a path file aswell

        Parameters
        ----------
        path: str
            Path to run this function into

        Returns
        -------
        dict[str, str]
            A dictionary that links a filename to its content, fileName : fileContent
        """

        path = os.path.abspath(path)

        if not os.path.exists(path):
            raise DSLParserPathNotFound(path)

        files = {}

        if os.path.isdir(path):
            for content in os.listdir(path):
                files.update(DSLParser.__getfiles(f'{path}/{content}'))

        if os.path.isfile(path):
            with open(path, 'r') as f:
                files.update({path: f.read()})

                f.close()

        return files

    def run(path: str) -> ParsedFile:
        """Run the DSLParser

        Parameters
        ----------
        path: str
            Path to run the parser into

        Returns
        -------
        ParsedFile
            A ParsedFile object with the content of all the files in the input path
        """

        try:
            files = DSLParser.__getfiles(path)
            lexer = Lexer(files)
            token_list = lexer.run()
            parser = TokenParser(token_list)
            ast = parser.run()

            interpreter = Interpreter()
            parsedFile = interpreter.run(ast)

            return parsedFile
        except DSLParserBaseException as e:
            raise e
        except Exception as e:
            raise e



from abc import ABC, abstractmethod

import thipster.engine.parsed_file as pf
import thipster.engine.resource_model as rm


class I_Terraform(ABC):
    """Terraform module interface
    """
    @abstractmethod
    def apply(self, plan_file_path: str | None = None):
        """Apply generated terraform code

        Parameters
        ----------
        plan_file_path : str, optional
            Path to plan file, by default None

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement apply()')

    @abstractmethod
    def generate(self, file: pf.ParsedFile, models: dict[str, rm.ResourceModel]):
        """Generates Terraform code from parsed file and models

        Parameters
        ----------
        file : pf.ParsedFile
            The ParsedFile object containing the resources defined in the input file
        models : dict[str, rm.ResourceModel]
            The dictionary of models

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement generate()')

    @abstractmethod
    def init(self):
        """Init Terraform for generated terraform code

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement generate()')

    @abstractmethod
    def plan(self):
        """Get plan from generated terraform code

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement plan()')



"""_Engine.py module.
"""

import thipster.engine.parsed_file as pf

from .i_auth import I_Auth
from .i_parser import I_Parser
from .i_repository import I_Repository
from .i_terraform import I_Terraform
from .resource_model import ResourceModel


class Engine():
    """The engine of thipster

    The core of the application, it is used to call and link all
    interfaces together.
    """

    def __init__(
            self, parser: I_Parser,
            repository: I_Repository,
            auth: I_Auth,
            terraform:  I_Terraform,
    ):
        """
        Parameters
        ----------
        parser : I_Parser
            Instance of a Parser class
        repository : I_Repository
            Instance of a Respository class
        auth : I_Auth
            Instance of an Auth class
        terraform : I_Terraform
            Instance of a Terraform class

        """
        self.__parser = parser
        self.__repository = repository
        self.__auth = auth
        self.__terraform = terraform

    @property
    def parser(self):
        return self.__parser

    @parser.setter
    def parser(self, value):
        if not isinstance(value, I_Parser):
            raise Exception()

        self.__parser = value

    @property
    def repository(self):
        return self.__repository

    @repository.setter
    def repository(self, value):
        if not isinstance(value, I_Repository):
            raise Exception()

        self.__repository = value

    @property
    def auth(self):
        return self.__auth

    @auth.setter
    def auth(self, value):
        if not isinstance(value, I_Auth):
            raise Exception()

        self.__auth = value

    @property
    def terraform(self):
        return self.__terraform

    @terraform.setter
    def terraform(self, value):
        if not isinstance(value, I_Terraform):
            raise Exception()

        self.__terraform = value

    def run(self, path: str) -> tuple[list[str], str]:
        """Returns json Terraform files from the input file name

        Calls the different run methods of the parser, repository,
        auth and terraform modules.
        Transforms the inputed filename into a Cloud architecture plan.

        Parameters
        ----------
        path : str
            The path of the files to be processed

        Returns
        -------
        tuple[list[str], str]
            A tuple made up of the list of directories containing the Terraform json
            files and a string with the results of the Terraform plan
        """
        # Parse file or directory
        parsed_file = self._parse_files(path)

        # Get needed models
        models = self._get_models(parsed_file)

        # Generate Terraform files
        self._generate_tf_files(parsed_file, models)

        self.__terraform.init()

        return self.__terraform.plan()

    def _parse_files(self, path: str) -> pf.ParsedFile:
        """Parse the input file or directory

        Parameters
        ----------
        path : str
            The path of the files to be processed

        Returns
        -------
        pf.ParsedFile
            The ParsedFile object containing the resources defined in the input file
        """
        parsed_file = self.__parser.run(path)
        assert type(parsed_file) == pf.ParsedFile

        return parsed_file

    def _get_models(
        self, file: pf.ParsedFile,
    ) -> dict[str, ResourceModel]:
        """Get the models from the repository

        Parameters
        ----------
        file : pf.ParsedFile
            The ParsedFile object containing the resources defined in the input file

        Returns
        -------
        dict[str, ResourceModel]
            The dictionary of models
        """
        types = [r.type for r in file.resources]
        models = self.__repository.get(types)

        return models

    def _generate_tf_files(
        self, file: pf.ParsedFile, models: dict[str, ResourceModel],
    ) -> list[str]:
        """Generate Terraform files

        Parameters
        ----------
        file : pf.ParsedFile
            The ParsedFile object containing the resources defined in the input file
        models : dict[str, ResourceModel]
            The dictionary of models

        Returns
        -------
        list[str]
            A list of directories containing the Terraform json files
        """
        self.__terraform.generate(file, models, self.__auth)

    def _init_terraform(self) -> None:
        """Initialize Terraform
        """
        self.__terraform.init()

    def _plan_terraform(self) -> str:
        """Plan Terraform

        Returns
        -------
        str
            The results of the Terraform plan
        """
        results = self.__terraform.plan()

        return results

    def _apply_terraform(self) -> str:
        """Apply Terraform

        Returns
        -------
        str
            The results of the Terraform apply
        """
        results = self.__terraform.apply()

        return results



from thipster.engine import THipsterException
from .token import TOKENTYPES as TT
from .token import Token


class DSLParserBaseException(THipsterException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class DSLParserPathNotFound(DSLParserBaseException):
    def __init__(self, file, *args: object) -> None:
        super().__init__(*args)
        self.file = file

    @property
    def message(self) -> str:
        return f'Path not found : {self.file}'


class DSLSyntaxException(DSLParserBaseException):
    def __init__(self, token: Token, expected: TT | list[TT], *args: object) -> None:
        super().__init__(*args)
        self.token = token
        self.expected = expected

    @property
    def message(self) -> str:
        if type(self.expected) is TT:
            return f"""{str(self.token.position)} :\n\tSyntax error : Expected \
{str(self.expected.value)}, got {str(self.token.tokenType)}"""
        else:
            return f"""{str(self.token.position)} :\n\tSyntax error : Expected \
{str(' or '.join(list(map(lambda x : str(x), self.expected))))}, got {
    str(self.token.tokenType)}"""


class DSLConditionException(DSLParserBaseException):
    def __init__(self, token: Token, *args: object) -> None:
        super().__init__(*args)
        self.token = token

    @property
    def message(self) -> str:
        return f"""{str(self.token.position)} :\n\tBad condition"""


class DSLUnexpectedEOF(DSLParserBaseException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

    @property
    def message(self) -> str:
        return 'Unexpected EOF'


class DSLParserNoEndingQuotes(DSLParserBaseException):
    def __init__(self, position, *args: object) -> None:
        super().__init__(*args)
        self.position = position

    @property
    def message(self) -> str:
        return f'Invalid syntax, missing ending quotes at : {self.position}'


class DSLParserVariableAlreadyUsed(DSLParserBaseException):
    def __init__(self, var: str, *args: object) -> None:
        super().__init__(*args)
        self.variable = var

    @property
    def message(self) -> str:
        return f'Variable already used : {self.variable}',


class DSLParserVariableNotDeclared(DSLParserBaseException):
    def __init__(self, var: str, *args: object) -> None:
        super().__init__(*args)
        self.variable = var

    @property
    def message(self) -> str:
        return f'Variable not declared : {self.variable}',



import os

import yaml
from jinja2 import Environment, FileSystemLoader
from thipster.engine import THipsterException

import thipster.engine.parsed_file as pf
from thipster.engine import I_Parser


class YAMLParserBaseException(THipsterException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class YAMLParserPathNotFound(YAMLParserBaseException):
    def __init__(self, file, *args: object) -> None:
        super().__init__(*args)
        self.file = file

    @property
    def message(self) -> str:
        return f'Path not found : {self.file}',


class YAMLParserNoName(YAMLParserBaseException):
    def __init__(self, resource, *args: object) -> None:
        super().__init__(*args)
        self.resource = resource

    @property
    def message(self) -> str:
        return f'No name for resource : {self.resource}'


class YAMLParser(I_Parser):

    def __getfiles(path: str) -> list[str]:
        """Recursively get all files names in the requested directory and its\
              sudirectories
        Can be run on a path file aswell

        Parameters
        ----------
        path: str
            Path to run this function into

        Returns
        -------
        list[str]
            A list of all the filenames
        """
        path = os.path.abspath(path)

        if not os.path.exists(path):
            raise YAMLParserPathNotFound(path)

        files = []

        if os.path.isdir(path):
            for content in os.listdir(path):
                files += YAMLParser.__getfiles(f'{path}/{content}')

        if os.path.isfile(path):
            return [path]

        return files

    def run(path: str) -> pf.ParsedFile:
        """Run the YAMLParser

        Parameters
        ----------
        path: str
            Path to run the parser into

        Returns
        -------
        ParsedFile
            A ParsedFile object with the content of all the files in the input path
        """
        try:
            files = YAMLParser.__getfiles(path)
            parsedFile = pf.ParsedFile()

            for file in files:
                filedir, filename = os.path.split(file)

                environment = Environment(loader=FileSystemLoader(filedir))
                template = environment.get_template(filename)
                rendered = template.render()
                content = yaml.safe_load(rendered)

                parsedFile.resources += YAMLParser.__convert(content)

            return parsedFile
        except yaml.YAMLError as exc:
            raise exc
        except YAMLParserBaseException as e:
            raise e
        except Exception as e:
            raise e

    def __convert(file: dict) -> list[pf.ParsedResource]:
        """Converts a dictionnary into a list of ParsedResources

        Parameters
        ----------
        file: dict
            Dict to convert

        Returns
        -------
        list[ParsedResource]
            Converted dict
        """
        resources = []

        for key, val in file.items():
            if type(val) == list:
                for res in val:
                    if 'name' in res:
                        name = res['name']
                        del res['name']
                    else:
                        raise YAMLParserNoName(key)

                    resources.append(
                        YAMLParser.__get_resource(
                            content=res, resourceType=key, name=name,
                        ),
                    )
            elif type(val) == dict:
                if 'name' in val:
                    name = val['name']
                    del val['name']
                else:
                    raise YAMLParserNoName(key)

                resources.append(
                    YAMLParser.__get_resource(
                        content=val, resourceType=key, name=name,
                    ),
                )

        return resources

    def __get_resource(content: dict, resourceType: str, name: str)\
            -> pf.ParsedResource:
        """Converts a dict in a ParsedResource

        Parameters
        ----------
        content: dict
            Dict of attributes of the resource
        resourceType: str
            Type of the resource
        name: str
            Name of the resource

        Returns
        -------
        ParsedResource
            A ParsedResource object with the content of the dict
        """
        attr = []

        for key, val in content.items():
            attr.append(YAMLParser.__get__attr(key, val))

        return pf.ParsedResource(
            type=resourceType,
            name=name,
            position=None,
            attributes=attr,
        )

    def __get__attr(name: str, value: object) -> pf.ParsedAttribute:
        """Converts an object in a ParsedAttribute

        Parameters
        ----------
        value: object
            Object to convert
        name: str
            Name of the attribute

        Returns
        -------
        ParsedResource
            A ParsedAttribute object with wanted value type
        """
        if type(value) == dict:
            val = YAMLParser.__get_dict(value)
        elif type(value) == list:
            val = YAMLParser.__get_list(value)
        else:
            val = pf.ParsedLiteral(value)

        return pf.ParsedAttribute(
            name=name,
            value=val,
            position=None,
        )

    def __get_dict(input: dict) -> pf.ParsedDict:
        """Converts a dict into a list of ParsedDict

        Parameters
        ----------
        input: dict
            Dict to convert

        Returns
        -------
        ParsedDict
            Converted dict
        """
        attr = []

        for key, val in input.items():
            attr.append(YAMLParser.__get__attr(key, val))

        return pf.ParsedDict(attr)

    def __get_list(input: list) -> pf.ParsedList:
        """Converts a dictionnary into a ParsedList

        Parameters
        ----------
        input: list
            List to convert

        Returns
        -------
        ParsedList
            Converted list
        """
        attr = []

        for val in input:
            if isinstance(val, dict):
                attr.append(YAMLParser.__get_dict(val))
            else:
                attr.append(pf.ParsedLiteral(val))

        return pf.ParsedList(attr)



from abc import ABC, abstractmethod

from thipster.engine.parsed_file import Position

from .token import Token


class Node(ABC):

    @abstractmethod
    def accept(self, visitor):
        raise NotImplementedError()


class StringNode(Node):
    def __init__(self, value: Token) -> None:
        super().__init__()
        self.__value = value

    def __str__(self) -> str:
        return f'<STRING {self.__value}>'

    @property
    def value(self) -> Token:
        return self.__value

    def accept(self, visitor):
        return visitor.visitString(self)


class StringExprNode(Node):
    def __init__(self, *values: Node) -> None:
        super().__init__()
        self.__value = []
        for v in values:
            if isinstance(v, list):
                self.__value.extend(v)
            else:
                self.__value.append(v)

    def __str__(self) -> str:
        return f"<STRING-EXPR {' '.join(list(map(lambda x : str(x), self.__value)))}>"

    @property
    def values(self) -> list[Node]:
        return self.__value

    def accept(self, visitor):
        return visitor.visitStringExpr(self)


class VariableNode(Node):
    def __init__(self, name: Token) -> None:
        super().__init__()
        self.__name = name

    def __str__(self) -> str:
        return f'<VAR {self.__name}>'

    @property
    def name(self):
        return self.__name.value

    def accept(self, visitor):
        return visitor.visitVariable(self)


class BoolNode(Node):
    def __init__(self, value: Token) -> None:
        super().__init__()
        self.__value = value

    def __str__(self) -> str:
        return f'<BOOL {self.__value}>'

    @property
    def value(self) -> bool:
        return bool(self.__value.value)

    def accept(self, visitor):
        return visitor.visitBool(self)


class IntNode(Node):
    def __init__(self, value: Token) -> None:
        super().__init__()
        self.__value = value

    def __str__(self) -> str:
        return f'<INT {self.__value}>'

    @property
    def value(self):
        return self.__value.value

    def accept(self, visitor):
        return visitor.visitInt(self)


class FloatNode(Node):
    def __init__(self, value: Token) -> None:
        super().__init__()
        self.__value = value

    def __str__(self) -> str:
        return f'<FLOAT {self.__value}>'

    @property
    def value(self):
        return self.__value.value

    def accept(self, visitor):
        return visitor.visitFloat(self)


class VariableDefinitionNode(Node):
    def __init__(self, name: Token, value: IntNode) -> None:
        super().__init__()
        self.__name = name
        self.__value = value

    def __str__(self) -> str:
        return f'<VARDEF {self.__name} = {self.__value}>'

    @property
    def name(self):
        return self.__name.value

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, __value):
        self.__value = __value

    def accept(self, visitor):
        return visitor.visitVariableDefinition(self)


class IfNode(Node):
    def __init__(self, condition: StringNode, ifCase: Node | None) -> None:
        super().__init__()
        self.__condition = condition
        self.__ifCase = ifCase

    def __str__(self) -> str:
        return f'<IF {self.__condition}: {self.__ifCase}>'

    @property
    def condition(self):
        return self.__condition

    @property
    def ifCase(self):
        return self.__ifCase

    @ifCase.setter
    def ifCase(self, __value):
        self.__ifCase = __value

    def accept(self, visitor):
        return visitor.visitIf(self)


class IfElseNode(IfNode):
    def __init__(
        self, condition: StringNode, ifCase: Node | None,
        elseCase: Node | None,
    ) -> None:
        super().__init__(condition, ifCase)
        self.__elseCase = elseCase

    def __str__(self) -> str:
        return f'<IF_E {str(self.condition)}: {str(self.ifCase)}, ELSE : \
{str(self.elseCase)}>'

    @property
    def elseCase(self):
        return self.__elseCase

    def accept(self, visitor):
        return visitor.visitIfElse(self)


class AmountNode(Node):
    def __init__(
        self,
        amount: IntNode, variable: VariableDefinitionNode | None, node: Node | None,
    )\
            -> None:
        super().__init__()
        self.__amount = amount
        self.__node = node
        self.__variable = variable

    def __str__(self) -> str:
        return f'<AMOUNT {self.__amount} #{self.__variable}: {self.__node}>'

    @property
    def node(self) -> Node | None:
        return self.__node

    @property
    def amount(self) -> Node:
        return self.__amount

    @property
    def variable(self) -> VariableDefinitionNode | None:
        return self.__variable

    @node.setter
    def node(self, __value):
        self.__node = __value

    def accept(self, visitor):
        return visitor.visitAmount(self)


class ValueNode(Node, ABC):

    @abstractmethod
    def accept(self, visitor):
        raise NotImplementedError()

    @property
    @abstractmethod
    def value(self):
        raise NotImplementedError()


class ParameterNode(Node):
    pass


class ParameterNode(Node):  # noqa: F811
    def __init__(self, name: StringNode, value: ValueNode) -> None:
        super().__init__()
        self.__name = name
        self.__value = value
        self.__position = None

    def __str__(self) -> str:
        return f'<PARAMETER name = {str(self.__name)}, value = {str(self.__value)}>'

    @property
    def name(self):
        return self.__name

    @property
    def value(self):
        return self.__value

    @property
    def position(self) -> Position | None:
        return self.__position

    @value.setter
    def value(self, __value):
        self.__value = __value

    def accept(self, visitor):
        return visitor.visitParameter(self)


class DictNode(ValueNode):
    def __init__(self, value: list[ParameterNode]) -> None:
        super().__init__()
        self.__value = value

    def __str__(self) -> str:
        return f"<DICT {' '.join(list(map(lambda x : str(x), self.__value)))}>"

    @property
    def value(self):
        return self.__value

    def accept(self, visitor):
        return visitor.visitDict(self)


class LiteralNode(ValueNode):
    def __init__(self, value: Node) -> None:
        super().__init__()
        self.__value = value

    def __str__(self) -> str:
        return f'<LITERAL {str(self.__value)}>'

    @property
    def value(self) -> Node:
        return self.__value

    def accept(self, visitor):
        return visitor.visitLiteral(self)


class ListNode(ValueNode):
    def __init__(self, value: list[ValueNode]) -> None:
        super().__init__()
        self.__value = value

    def __str__(self) -> str:
        return f"<LIST {' '.join(list(map(lambda x : str(x), self.__value)))}>"

    @property
    def value(self):
        return self.__value

    def accept(self, visitor):
        return visitor.visitList(self)


class CompExprNode(Node):
    def __init__(self, leftValue, operation, rightValue=None) -> None:
        super().__init__()
        self.__leftValue = leftValue
        self.__operation = operation
        self.__rightValue = rightValue

    @property
    def leftValue(self) -> Node:
        return self.__leftValue

    @property
    def operation(self) -> Node:
        return self.__operation

    @property
    def rightValue(self) -> Node:
        return self.__rightValue

    def accept(self, visitor):
        return visitor.visitCompExpr(self)


class ArithExprNode(Node):
    def __init__(self, terms, operations) -> None:
        super().__init__()
        if len(terms) != len(operations) + 1:
            raise
        self.__terms = terms
        self.__operations = operations

    @property
    def terms(self) -> Node:
        return self.__terms

    @property
    def operation(self) -> Node:
        return self.__operations

    def accept(self, visitor):
        return visitor.visitArithExpr(self)


class TermNode(Node):
    def __init__(self, factors, operations) -> None:
        super().__init__()
        if len(factors) != len(operations) + 1:
            raise
        self.__factors = factors
        self.__operations = operations

    @property
    def factors(self) -> Node:
        return self.__factors

    @property
    def operation(self) -> Node:
        return self.__operations

    def accept(self, visitor):
        return visitor.visitTerm(self)


class FactorNode(Node):
    def __init__(self, factors, operation) -> None:
        super().__init__()
        self.__factors = factors
        self.__operations = operation

    @property
    def factors(self) -> Node:
        return self.__factors

    @property
    def operation(self) -> Node:
        return self.__operations

    def accept(self, visitor):
        return visitor.visitFactor(self)


class ResourceNode(Node):
    def __init__(
        self,
        resourceType: StringNode,
        name: StringNode,
        parameters: DictNode | ListNode,
    ) -> None:
        super().__init__()
        self.__type = resourceType
        self.__name = name
        self.__position = None
        self.__parameters = parameters

    def __str__(self) -> str:
        return f'<RESOURCE type = {str(self.__type)}, \
name = {str(self.__name)}, parameters = {str(self.__parameters)}>'

    @property
    def type(self) -> StringNode:
        return self.__type

    @property
    def name(self) -> StringNode:
        return self.__name

    @property
    def position(self) -> Position | None:
        return self.__position

    @property
    def parameters(self) -> DictNode | ListNode:
        return self.__parameters

    def accept(self, visitor):
        return visitor.visitResource(self)


class FileNode(Node):
    def __init__(self) -> None:
        super().__init__()
        self.__resources = []

    def add(self, item: ResourceNode | IfNode | AmountNode) -> None:
        self.__resources.append(item)

    def __str__(self) -> str:
        return '\n'.join(list(map(lambda x: str(x), self.__resources)))

    @property
    def resources(self) -> list[ResourceNode]:
        return self.__resources

    def accept(self, visitor):
        return visitor.visitFile(self)



import copy
import importlib
import os
import shutil
import subprocess
import sys
import uuid

from cdktf import App, TerraformStack
from constructs import Construct
from python_terraform import Terraform

import thipster.engine.parsed_file as pf
import thipster.engine.resource_model as rm
import thipster.terraform.exceptions as cdk_exceptions
from thipster.engine import I_Auth, I_Terraform
from thipster.helpers import create_logger as Logger


class CDK(I_Terraform):
    _models = []
    _parent_resources_stack = []

    _inherited_attributes: list[pf.ParsedAttribute] = []
    _created_resources = {}
    _logger = Logger(__name__)

    def apply(self, plan_file_path: str | None = None):
        """Applies generated Terraform plan

        Parameters
        ----------
        plan_file_path : str, optional
            path to the plan file, default None

        Returns
        -------
        str
            Terraform apply output
        """
        t = Terraform()
        _, stdout, stderr = t.apply(plan_file_path)
        return stdout + stderr

    def generate(
        self, file: pf.ParsedFile, models: dict[str, rm.ResourceModel],
        _auth: I_Auth,
    ):
        """Generate Terraform file from given parsed file and models

        Parameters
        ----------
        file : pf.ParsedFile
            parsedFile object to transform
        models : dict[str, rm.ResourceModel]
            associated models
        _auth : I_Auth
            authenticator to use
        """
        CDK._created_resources = {}
        CDK._models = models
        CDK._parent_resources_stack = []

        # Init CDK
        app = App()

        f_position = file.resources[0].position
        file_name = f_position.fileName if f_position else 'thipster_infrastructure'

        CDK._logger.debug('Creating tf code for file %s', file_name)

        # Declare new stack in CDK
        class _ResourceStack(TerraformStack):
            def __init__(self, scope: Construct, ns: str):
                super().__init__(scope, ns)

                _auth.authenticate(self)

                for resource in file.resources:
                    res = CDK._create_resource_from_resource(
                        self,
                        resource=resource,
                    )

                    CDK._created_resources[f'{resource.type}/{resource.name}'] = res.id

        _ResourceStack(app, file_name)

        # Synth files
        app.synth()

        self.__dirs = [
            f'{app.outdir}/stacks/{c.node.path}' for c in app.node.children
        ]

        CDK._logger.info('Created %s terraform file(s)', len(self.__dirs))

        # Move files
        for dirname in self.__dirs:
            shutil.move(
                os.path.join(os.getcwd(), dirname, 'cdk.tf.json'),
                os.path.join(os.getcwd(), 'thipster.tf.json'),
            )

        # Delete cdktf.out directory
            for content in os.listdir(os.path.join(os.getcwd(), dirname)):
                os.remove(f'{dirname}/{content}')
            os.rmdir(dirname)
        for content in os.listdir(os.path.join(os.getcwd(), 'cdktf.out')):
            d = f'cdktf.out/{content}'
            os.rmdir(d) if os.path.isdir(d) else os.remove(d)
        os.rmdir('cdktf.out')

    def init(self):
        """Initilize terraform

        Returns
        -------
        str
            Terraform init output
        """
        t = Terraform()
        _, stdout, stderr = t.init()
        return stdout + stderr

    def plan(self):
        """Get plan from generated terraform code

        Returns
        -------
        str
            Terraform plan output
        """
        t = Terraform()
        _, stdout, stderr = t.plan(out='thipster.tfplan')
        return stdout + stderr

    def _pip_install(package: str):
        """Install a package if it wasn't already installed by thipster

        Parameters
        ----------
        package : str
            package to install
        """
        if package not in sys.modules:
            subprocess.check_call(
                [sys.executable, '-m', 'pip', 'install', '-qqq', package],
            )

    def _import(package_name: str, module_name: str, class_name: str) -> type:
        """Import a class from any package

        Parameters
        ----------
        package_name : str
            package that contains the class
        module_name : str
            module that contains the class
        class_name : str
            class to import

        Returns
        -------
        type :
            the imported class
        """
        module = importlib.import_module(f'{package_name}.{module_name}')
        class_ = getattr(module, class_name)

        return class_

    def _create_default_resource(
        self, resource_type: str, parent_name: str | None = None,
        no_modif: bool = True, no_dependencies: bool = False,
    ):
        """Create a resource with all default values

        Parameters
        ----------
        self : _ResourceStack
            Terraform CDK stack that contains the resource
        resource_type : str
            type of the resource to create
        parent_name : str, optional
            name of its parent, default None
        no_modif : bool, optional
            if True, raise errors if resource can't be created with default values, \
default False
        no_dependencies : bool, optional
            if True, create the default dependencies, default False

        Returns
        -------
        (type, str, dict[str, object]) :
            all the information needed to instantiate the class with default values
        """
        # Checks for cyclic dependency
        if resource_type in CDK._parent_resources_stack:
            CDK._logger.error(
                '%s already present in parent Stack', resource_type,
            )
            raise cdk_exceptions.CDKCyclicDependencies(
                CDK._parent_resources_stack,
            )

        CDK._parent_resources_stack.append(resource_type)

        model = CDK._models[resource_type]

        attributes = copy.deepcopy(model.attributes)

        for a in CDK._inherited_attributes:
            if a.name in attributes.keys():
                attributes[a.name].default = rm.Model_Literal(a.value)

            else:
                attributes[a.name] = rm.Model_Attribute(
                    cdk_name=a.name,
                    default=rm.Model_Literal(a.value),
                )

        # Checks that all attributes have a default value
        if (
            no_modif
            and not all(map(lambda x: x.default is not None, attributes.values()))
        ):
            raise cdk_exceptions.CDKMissingAttributeInDependency(resource_type)

        # Import package and class
        CDK._pip_install(model.cdk_provider)
        resource_class = CDK._import(
            model.cdk_provider, model.cdk_module, model.cdk_name,
        )

        # Process name + default values
        dependencies = copy.deepcopy(model.dependencies)
        resource_args = {}

        name = f'{parent_name}-{uuid.uuid4()}'
        if model.name_key:
            resource_args[model.name_key] = name

        for attribute_name, attribute_value in attributes.items():

            if not no_dependencies and attribute_name in dependencies:

                CDK._check_explicit_dependency(
                    self, resource_args, dependencies[attribute_name]['resource'],
                    attribute_value.default, attribute_name,
                )

                del dependencies[attribute_name]

            if attribute_name in model.attributes:
                resource_args[attribute_value.cdk_name] = attribute_value.default

        # Create default defendencies if needed
        if not no_dependencies:
            CDK._generate_default_dependencies(
                self, dependencies, model, resource_args, parent_name,
            )

        CDK._logger.debug('Created default %s named %s', resource_class, name)

        return (resource_class, name, resource_args)

    def _create_resource(self, resource_args: object, resource_type: str = None):
        """Create a resource with the given values

        Parameters
        ----------
        self : _ResourceStack
            Terraform CDK stack that contains the resource
        resource_args : object
            data source to create the resource
        resource_type : str, optional
            type of the resource to create, default None

        Returns
        -------
        object :
            the created resource
        """

        if isinstance(resource_args, pf.ParsedResource):
            return CDK._create_resource_from_resource(self, resource_args)

        return CDK._create_resource_from_args(self, resource_type, resource_args)

    def _create_resource_from_args(
        self, resource_type: str, input_args: list[pf.ParsedAttribute] | None,
    ):
        """Create a resource from a list of ParsedAttributes

        Parameters
        ----------
        self : _ResourceStack
            Terraform CDK stack that contains the resource
        resource_type : str, optional
            type of the resource to create, default None
        input_args : list[ParsedAttribute]
            data source to create the resource

        Returns
        -------
        object :
            the created resource
        """
        resource_class, resource_name, resource_args = CDK._create_default_resource(
            self,
            resource_type,
            no_modif=False,
            no_dependencies=True,
        )
        model = CDK._models[resource_type]

        # Process attributes
        dependencies = copy.deepcopy(model.dependencies)

        def attributes(attribute_list):
            for attribute in attribute_list:
                if attribute.name and model.name_key:
                    resource_args[model.name_key] = attribute.name

                CDK._process_attribute(
                    self, model, attribute, resource_args, dependencies,
                )

        attributes(input_args)
        attributes(CDK._inherited_attributes)

        CDK._inherited_attributes += input_args
        CDK._generate_default_dependencies(
            self, dependencies, model, resource_args, resource_name,
        )
        CDK._inherited_attributes = CDK._inherited_attributes[
            :-len(
                input_args,
            )
        ]

        # Create resource
        CDK._parent_resources_stack.remove(resource_type)

        CDK._logger.debug(
            'Created default %s named %s',
            resource_class, resource_name,
        )

        if not model.name_key:
            return resource_class(**resource_args)
        return resource_class(self, resource_name, **resource_args)

    def _create_resource_from_resource(self, resource: pf.ParsedResource):
        # Create resource with default values
        resource_class, _, resource_args = CDK._create_default_resource(
            self,
            resource.type,
            no_modif=False,
            no_dependencies=True,
        )
        """Create a resource from a list of ParsedAttributes

        Parameters
        ----------
        self : _ResourceStack
            Terraform CDK stack that contains the resource
        resource : ParsedResource
            data source to create the resource

        Returns
        -------
        object :
            the created resource
        """
        model = CDK._models[resource.type]

        if model.name_key:
            resource_args[model.name_key] = resource.name

        # Process attributes
        dependencies = copy.deepcopy(model.dependencies)

        def attributes(attribute_list):
            for attribute in attribute_list:
                CDK._process_attribute(
                    self,
                    model,
                    attribute,
                    resource_args,
                    dependencies,
                )

        attributes(resource.attributes)
        attributes(CDK._inherited_attributes)

        CDK._inherited_attributes += resource.attributes
        CDK._generate_default_dependencies(
            self, dependencies, model, resource_args, resource.name,
        )
        CDK._inherited_attributes = CDK._inherited_attributes[
            :-len(
                resource.attributes,
            )
        ]

        # Create resource
        CDK._parent_resources_stack.remove(resource.type)

        CDK._logger.debug(
            'Created resource %s named %s',
            resource_class, resource.name,
        )

        return resource_class(self, resource.name, **resource_args)

    def _process_attribute(
        self,
        model: rm.ResourceModel,
        attribute: pf.ParsedAttribute,
        resource_args: dict[str, object],
        dependencies: dict[str, dict[str, object]] | None,
    ):
        """Process an attribute

        Parameters
        ----------
        self : _ResourceStack
            Terraform CDK stack that contains the resource
        model : ResourceModel
            model of the resource that is being created
        attribute : ParsedAttribute
            attribute to handle
        resource_args : object
            data source needed to create the resource
        dependencies : dict[str, dict[str, object]]
            dependencies needed to create the resource
        """
        if attribute.name in dependencies:
            CDK._check_explicit_dependency(
                self, resource_args, dependencies[attribute.name]['resource'],
                attribute.value, attribute.name,
            )
            del dependencies[attribute.name]
            return

        # Checks if attribute is an internal object
        if attribute.name in model.internal_objects:
            resource_args = CDK._create_internal_object(
                self,
                attribute.name,
                model.internal_objects[attribute.name],
                attribute.value,
                resource_args,
            )
            return

        # Checks if attribute is expected as an attibute
        if attribute.name not in model.attributes:
            return

        # Processes list attribute
        attribute_value = attribute.value
        if model.attributes[attribute.name].is_list:
            if type(attribute.value) is list:
                attribute_value = [i.value for i in attribute.value]
            else:
                attribute_value = [attribute.value]

        # Sets attribute value
        resource_args[model.attributes[attribute.name].cdk_name] = attribute_value

    def _generate_default_dependencies(
            self,
            dependencies,
            model: rm.ResourceModel,
            resource_args,
            resource_name,
    ):
        """Generate default dependencies and internal objects in a resource

        Parameters
        ----------
        self : _ResourceStack
            Terraform CDK stack that contains the resource
        dependencies :
            list of dependencies to create
        model : str
            model of the resource that is being created
        resource_args : _type_
            data source needed to create the resource
        resource_name : dict
            name of the resource that is being created
        """
        # Create missing dependencies
        for dependency_name, dependency_object in dependencies.items():
            CDK._create_dependency(
                self, dependency_name, dependency_object, resource_args, resource_name,
            )

        # Create default internal objects if needed
        for internal_object_name, internal_object in model.internal_objects.items():
            if internal_object_name not in resource_args:
                for default_args_json in internal_object['defaults']:

                    # Transform in list of ParsedAttributes
                    default_args = []
                    for arg_key, arg_value in default_args_json.items():
                        default_args.append(
                            pf.ParsedAttribute(
                                arg_key, None, pf.ParsedLiteral(arg_value),
                            ),
                        )

                    CDK._create_internal_object(
                        self,
                        internal_object_name,
                        model.internal_objects[internal_object_name],
                        default_args,
                        resource_args,
                    )

    def _create_internal_object(
        self,
        name: str,
        internal_object_model,
        internal_object_args,
        resource_args: dict,
    ):
        """Create an internal object in a resource

        Parameters
        ----------
        self : _ResourceStack
            Terraform CDK stack that contains the resource
        name : str
            name of the internal object
        internal_object_model : ResourceModel
            model of the internal object
        internal_object_args : object
            data source to create the internal object
        resource_args : dict
            data source needed to create the resource
        """
        internal_object_type = internal_object_model['var_type'] \
            if 'var_type' in internal_object_model else 'Unknown'

        if internal_object_type.startswith('list'):
            if name not in resource_args:
                resource_args[name] = []

            if isinstance(internal_object_args, list) \
                    and isinstance(internal_object_args[0], pf.ParsedDict):

                for internalObject in internal_object_args:
                    res = CDK._create_resource(
                        self,
                        internalObject.value,
                        internal_object_model['resource'],
                    )
                    resource_args[name] += [res]

                return

            res = CDK._create_resource(
                self,
                internal_object_args,
                internal_object_model['resource'],
            )
            resource_args[name] += [res]
            return

        res = CDK._create_resource(
            self,
            internal_object_args,
            internal_object_model['resource'],
        )
        resource_args[name] = res

        return

    def _create_dependency(self, dep_name, dep_value, resource_args, parent_name):
        """Create an internal object in a resource

        Parameters
        ----------
        self : _ResourceStack
            Terraform CDK stack that contains the resource
        dep_name : str
            name of the dependency
        dep_value : ResourceModel
            model of the dependency
        resource_args : dict
            data source needed to create the resource
        parent_name : str
            name of its parent
        """
        class_, class_name, class_attributes = CDK._create_default_resource(
            self, dep_value['resource'], parent_name,
        )

        id = class_(self, class_name, **class_attributes).id
        if dep_name:
            resource_args[dep_name] = id

        CDK._parent_resources_stack.remove(dep_value['resource'])

    def _check_explicit_dependency(
        self, resource_args, dependency_type, dependency_value, attribute_name,
    ):
        created_name = f'{dependency_type}/'\
            f'{dependency_value}'

        # Checks if attribute is an explicit dependency
        if created_name not in CDK._created_resources.keys():

            if isinstance(dependency_value, str):
                raise cdk_exceptions.CDKDependencyNotDeclared(
                    attribute_name, dependency_value,
                )

            # Creates explicit dependency
            CDK._create_dependency(
                self, attribute_name, dependency_value,
                resource_args, attribute_name,
            )

        resource_args[attribute_name] = CDK._created_resources[created_name]



import os

import pytest

import thipster.engine as eng
from thipster.engine.parsed_file import ParsedFile
from thipster.engine.resource_model import ResourceModel


def logger(name: str):
    def wrapper(function):
        def internal_wrapper(*args, **kwargs):
            print(f'{name} starting')
            res = function(*args, **kwargs)
            print(f'{name} returned :\n{str(res)}')
            return res
        return internal_wrapper
    return wrapper


class MockException(Exception):
    def __init__(self, message):
        self.message = message


class MockAuth(eng.I_Auth):
    @logger('- Authentifier')
    def authenticate(self):
        return (None, None)


class MockParser(eng.I_Parser):
    @logger('- Parser')
    def run(self, filename) -> str:
        return ParsedFile()


class MockRepository(eng.I_Repository):
    @logger('- Repo')
    def get(self, resourceNames: list[str]) -> dict[str, ResourceModel]:
        return []


class MockTerraform(eng.I_Terraform):
    @logger('- Terraform:apply')
    def apply(self):
        pass

    @logger('- Terraform:generate')
    def generate(self, a, b, auth):
        pass

    @logger('- Terraform:init')
    def init(self):
        pass

    @logger('- Terraform:plan')
    def plan(self):
        pass


def create_dir(dirname: str, files: dict[str, str]):
    if not os.path.isdir(dirname):
        os.mkdir(dirname)

    dirname = os.path.abspath(dirname)
    for name, content in files.items():
        create_file(name, content, dirname)

    def destroy_files():
        for content in os.listdir(dirname):
            os.remove(f'{dirname}/{content}')
        os.rmdir(dirname)

    return destroy_files


def create_file(filename: str, content: str, dirname: str = 'test'):
    if not os.path.isdir(dirname):
        os.mkdir(dirname)
    dirname = os.path.abspath(dirname)

    file = open(f'{dirname}/{filename}', 'w')
    file.write(content)
    file.close()


def test_engine_calls():
    parser = MockParser()
    repository = MockRepository()
    auth = MockAuth()
    terraform = MockTerraform()

    engine = eng.Engine(parser, repository, auth, terraform)
    res = engine.run('test.file')

    assert res is None


def test_parser_failure(mocker):
    def parserFail(self, filename: str):
        raise MockException('Parser failure')

    mocker.patch(
        'tests.engine.test_engine.MockParser.run',
        parserFail,
    )

    parser = MockParser()
    repository = MockRepository()
    auth = MockAuth()
    terraform = MockTerraform()

    with pytest.raises(MockException):
        engine = eng.Engine(parser, repository, auth, terraform)
        engine.run('test.file')


def test_repository_failure(mocker):
    def repositoryFail(self, resourceNames: list[str]) -> dict[str, ResourceModel]:
        raise MockException('Repository failure')

    mocker.patch(
        'tests.engine.test_engine.MockRepository.get',
        repositoryFail,
    )

    parser = MockParser()
    repository = MockRepository()
    auth = MockAuth()
    terraform = MockTerraform()

    with pytest.raises(MockException):
        engine = eng.Engine(parser, repository, auth, terraform)
        engine.run('test.file')



import os

from thipster.engine import I_Parser
from thipster.engine import THipsterException
from thipster.engine.parsed_file import ParsedFile
from .dsl_parser import DSLParser
from .yaml_parser import YAMLParser


class ParserPathNotFound(THipsterException):
    def __init__(self, path, *args: object) -> None:
        super().__init__(*args)

        self.__path = path

    @property
    def message(self) -> str:
        return f'Path not found : {self.__path}'


class ParserFactory(I_Parser):

    __parsers = {
        '.yaml': YAMLParser,
        '.yml': YAMLParser,
        '.jinja': YAMLParser,
        '.thips': DSLParser,
    }

    def addParser(parser: I_Parser, extensions: list[str]):
        ParserFactory.__parsers.update({e: parser for e in extensions})

    def __getfiles(self, path: str) -> list[str]:
        """Recursively get all files names in the requested directory and its\
              sudirectories
        Can be run on a path file aswell

        Parameters
        ----------
        path: str
            Path to run this function into

        Returns
        -------
        list[str]
            A list of all the filenames
        """

        path = os.path.abspath(path)

        if not os.path.exists(path):
            raise ParserPathNotFound(path)

        files = []

        if os.path.isdir(path):
            for content in os.listdir(path):
                files += self.__getfiles(f'{path}/{content}')

        if os.path.isfile(path):
            return [path]

        return files

    def __noParser(self, pathExtension):
        class noParser():
            def run(path):
                raise Exception(f'{pathExtension} files can\'t be parsed')

        return noParser

    def run(self, path: str) -> ParsedFile:
        """Run the ParserFactory

        Parameters
        ----------
        path: str
            Path to run the parser into

        Returns
        -------
        ParsedFile
            A ParsedFile object with the content of all the files in the input path
        """
        files = self.__getfiles(path)

        res = ParsedFile()
        for file in files:
            parsedFile = self.__getParser(file).run(file)
            res.resources += parsedFile.resources

        return res

    def __getParser(self, path) -> I_Parser:

        _, pathExtension = os.path.splitext(path)

        return ParserFactory.__parsers.get(
            pathExtension, self.__noParser(pathExtension),
        )



"""ResourceModel.py module.
"""

from abc import ABC, abstractmethod


class I_Model_Value(ABC):
    """Model Attribute Value Interface
    """

    @property
    @abstractmethod
    def value(self):
        pass


class Model_Attribute():
    """Represents a Resource Model attribute
    """

    def __init__(
            self, cdk_name: str,
            default: I_Model_Value | None = None,
            optional: bool = True,
            is_list: bool = False,
    ):
        """
        Parameters
        ----------
        name : str
            Attribute name
        default : I_Model_Value, optional
            Default Attribute value if there is one, by default None
        optional : bool, optional
            Is attribute optional ?, by default True
        is_list : bool, optional
            Is attribute a list ?, by default False
        """
        self.__cdk_name = cdk_name
        self.__default = default
        self.__optional = optional
        self.__is_list = is_list

    @property
    def default(self):
        return self.__default.value if self.__default else None

    @default.setter
    def default(self, value):
        self.__default = value

    @property
    def cdk_name(self):
        return self.__cdk_name

    @property
    def optional(self):
        return self.__optional

    @property
    def is_list(self):
        return self.__is_list


class Model_List(I_Model_Value):
    """Represents a List of values for a Resource Model attribute
    """

    def __init__(self, value: list[I_Model_Value | None] | None):
        super().__init__()
        self.__value = value

    @property
    def value(self):
        return self.__value

    def __iter__(self):
        self.i = 0
        return self

    def __next__(self):
        if not self.__value:
            return None
        ret = self.__value[self.i]
        if self.i > len(self.__value):
            raise StopIteration
        else:
            self.i += 1

        return ret


class Model_Literal(I_Model_Value):
    """Represents a literal value for a Resource Model attribute
    """

    def __init__(self, value):
        super().__init__()
        self.__value = value

    @property
    def value(self):
        return self.__value


class Model_Dict(I_Model_Value):
    """Represents a dictionary value for a Resource Model attribute
    """

    def __init__(self, value: dict[str, Model_Attribute] | None):
        super().__init__()
        self.__value = value

    @property
    def value(self):
        return self.__value

    def __iter__(self):
        self.i = 0
        return self

    def __next__(self):
        if not self.__value:
            return None
        ret = self.__value[self.i]
        if self.i > len(self.__value):
            raise StopIteration
        else:
            self.i += 1

        return ret


class ResourceModel():
    """Represents a Resource Model
    """

    def __init__(
            self,
            type: str,
            attributes: dict[str, Model_Attribute] | None,
            dependencies: dict[str, dict[str, object]] | None,
            internalObjects: dict[str, dict[str, object]] | None,
            name_key: str | None,
            cdk_provider: str,
            cdk_module: str,
            cdk_name: str,
    ):
        self.__type = type
        self.__attributes = attributes
        self.__dependencies = dependencies
        self.__internal_objects = internalObjects
        self.__name_key = name_key
        self.__cdk_provider = cdk_provider
        self.__cdk_module = cdk_module
        self.__cdk_name = cdk_name

    @property
    def type(self):
        return self.__type

    @property
    def attributes(self):
        return self.__attributes

    @property
    def dependencies(self):
        return self.__dependencies

    @property
    def internal_objects(self):
        return self.__internal_objects

    @property
    def name_key(self):
        return self.__name_key

    @property
    def cdk_provider(self):
        return self.__cdk_provider

    @property
    def cdk_module(self):
        return self.__cdk_module

    @property
    def cdk_name(self):
        return self.__cdk_name



from thipster.engine.parsed_file import Position
from thipster.parser.dsl_parser.token import Token


def test_create_token():
    position = Position(fileName='testFile', ln=2, col=6)
    tokenType = 'test_type'
    value = 'test_value'

    token = Token(position, tokenType=tokenType, value=value)

    assert isinstance(token, Token)
    assert token.position == position
    assert token.tokenType == tokenType
    assert token.value == value


def test_token_to_string():
    position = Position(fileName='testFile', ln=2, col=6)
    positionStr = '(File : testFile, Ln 2, Col 6)'
    assert str(position) == positionStr

    tokenType = 'test_type'
    value = 'test_value'
    token = Token(position, tokenType=tokenType, value=value)
    tokenStr = f'(Type: {str(tokenType)}, Position: {positionStr}, Value: {value})'
    assert repr(token) == tokenStr



import pytest

from thipster.parser.dsl_parser.exceptions import DSLParserNoEndingQuotes
from thipster.parser.dsl_parser.lexer import Lexer
from thipster.parser.dsl_parser.token import TOKENTYPES as TT


def __getTokenString(
    fileName: str,
    ln: int,
    col: int,
    tokenType: TT,
    value: str | None = None,
):
    positionStr = f'(File : {fileName}, Ln {ln}, Col {col})'
    tokenStr = f'(Type: {tokenType.value}, Position: {positionStr}'
    if value:
        tokenStr += f', Value: {value}'
    tokenStr += ')'
    return tokenStr


def __single_token_test(
    input: str,
    token: str,
    tokenValue: str | None = None,
    isNewLine: bool = False,
):

    lineValue, columnValue = 1, len(input)+1
    if isNewLine:
        lineValue, columnValue = 2, 1
    expectedOutput = [
        __getTokenString('test_file.thips', 1, 1, token, tokenValue),
        __getTokenString(
            'test_file.thips', lineValue,
            columnValue, TT.NEWLINE,
        ),
        __getTokenString('test_file.thips', lineValue+1, 1, TT.EOF),
    ]
    lexer = Lexer({'test_file.thips': input})
    lexer.run()
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_create_lexer():
    files = {
        'file_name_1': 'file_content_1',
        'file_name_2': 'file_content_2',
    }
    lexer = Lexer(files)

    assert isinstance(lexer, Lexer)
    assert lexer.files == files
    assert len(lexer.tokenList) == 0


def test_lex_single_tokens():
    for input, output in {
        ':': TT.COLON,
        ',': TT.COMMA,
        '+': TT.PLUS,
        '-': TT.MINUS,
        '=': TT.EQ,
        '*': TT.MUL,
        '/': TT.DIV,
        '!': TT.EXCLAMATION,
        '<': TT.LT,
        '>': TT.GT,
        '^': TT.POW,
    }.items():
        __single_token_test(input, output)


def test_lex_brackets():
    lexer = Lexer({'': ''})
    input = '[toto]'
    expectedOutput = [
        __getTokenString('file', 1, 1, TT.BRACKETS_START),
        __getTokenString('file', 1, 2, TT.STRING, 'toto'),
        __getTokenString('file', 1, 6, TT.BRACKETS_END),
        __getTokenString('file', 1, 7, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.EOF),
    ]
    lexer.lex(input, 'file')
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_lex_quoted_string():
    lexer = Lexer({'': ''})

    input = '"bucket number 21""bucket number 22"'
    expectedOutput = [
        __getTokenString('file', 1, 2, TT.STRING, 'bucket number 21'),
        __getTokenString('file', 1, 20, TT.STRING, 'bucket number 22'),
        __getTokenString('file', 1, 37, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.EOF),
    ]
    lexer.lex(input, 'file')
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]

    lexer = Lexer({'': ''})
    input = "'bucket number 21'"
    expectedOutput = [
        __getTokenString('file', 1, 2, TT.STRING, 'bucket number 21'),
        __getTokenString('file', 1, 19, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.EOF),
    ]
    lexer.lex(input, 'file')
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]

    lexer = Lexer({'': ''})
    input = "'bucket number \"21\"'"
    expectedOutput = [
        __getTokenString('file', 1, 2, TT.STRING, 'bucket number "21"'),
        __getTokenString('file', 1, 21, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.EOF),
    ]
    lexer.lex(input, 'file')
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]

    lexer = Lexer({'': ''})
    input = '"bucket number \'21\'"'
    expectedOutput = [
        __getTokenString('file', 1, 2, TT.STRING, "bucket number '21'"),
        __getTokenString('file', 1, 21, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.EOF),
    ]
    lexer.lex(input, 'file')
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_lex_quoted_string_error():
    input = {
        'file': 'bucket nom-8: "amount: 5 \n\t',
    }
    lexer = Lexer(input)
    with pytest.raises(DSLParserNoEndingQuotes):
        lexer.run()


def test_lex_multiline_input():
    input = {
        'file': 'bucket nom\\\ntest:',
    }
    expectedOutput = [
        __getTokenString('file', 1, 1, TT.STRING, 'bucket'),
        __getTokenString('file', 1, 7, TT.WHITESPACE),
        __getTokenString('file', 1, 8, TT.STRING, 'nomtest'),
        __getTokenString('file', 2, 5, TT.COLON),
        __getTokenString('file', 2, 6, TT.NEWLINE),
        __getTokenString('file', 3, 1, TT.EOF),
    ]
    lexer = Lexer(input)
    lexer.run()
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_lex_var():
    input = '#variable'
    expectedOutput = [
        __getTokenString('file', 1, 2, TT.VAR, 'variable'),
        __getTokenString('file', 1, 10, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.EOF),
    ]
    lexer = Lexer({'': ''})
    lexer.lex(input, 'file')
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_lex_newline():
    __single_token_test('\n', TT.NEWLINE, isNewLine=True)


def test_lex_tab():
    __single_token_test('\t', TT.TAB)


def test_lex_4_whitespaces_as_tab():
    input = {
        'file': '    \n\t',
    }
    expectedOutput = [
        __getTokenString('file', 1, 1, TT.TAB),
        __getTokenString('file', 1, 5, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.TAB),
        __getTokenString('file', 2, 2, TT.NEWLINE),
        __getTokenString('file', 3, 1, TT.EOF),
    ]
    lexer = Lexer(input)
    output = lexer.run()

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_lex_2_4_whitespaces_as_tabs():
    input = {
        'file': '        \n\t',
    }
    expectedOutput = [
        __getTokenString('file', 1, 1, TT.TAB),
        __getTokenString('file', 1, 5, TT.TAB),
        __getTokenString('file', 1, 9, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.TAB),
        __getTokenString('file', 2, 2, TT.NEWLINE),
        __getTokenString('file', 3, 1, TT.EOF),
    ]
    lexer = Lexer(input)
    output = lexer.run()

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_lex_dash():
    __single_token_test('-', TT.MINUS)


def test_lex_amount():
    input = 'amount '
    expectedOutput = [
        __getTokenString('file', 1, 1, TT.AMOUNT),
        __getTokenString('file', 1, 7, TT.WHITESPACE),
        __getTokenString('file', 1, 8, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.EOF),
    ]
    lexer = Lexer({'': ''})
    lexer.lex(input, 'file')
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_lex_and():
    __single_token_test('and', TT.AND)


def test_lex_if():
    __single_token_test('if', TT.IF)


def test_lex_elif():
    __single_token_test('elif', TT.ELIF)


def test_lex_else():
    __single_token_test('else', TT.ELSE)


def test_lex_if_else():
    input = 'if condition else something'
    expectedOutput = [
        __getTokenString('file', 1, 1, TT.IF),
        __getTokenString('file', 1, 3, TT.WHITESPACE),
        __getTokenString('file', 1, 4, TT.STRING, 'condition'),
        __getTokenString('file', 1, 13, TT.WHITESPACE),
        __getTokenString('file', 1, 14, TT.ELSE),
        __getTokenString('file', 1, 18, TT.WHITESPACE),
        __getTokenString('file', 1, 19, TT.STRING, 'something'),
        __getTokenString('file', 1, 28, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.EOF),
    ]
    lexer = Lexer({'': ''})
    lexer.lex(input, 'file')
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_lex_or():
    __single_token_test('or', TT.OR)


def test_lex_true():
    __single_token_test('true', TT.BOOLEAN, tokenValue='true')


def test_lex_false():
    __single_token_test('false', TT.BOOLEAN, tokenValue='false')


def test_lex_int():
    __single_token_test('8', TT.INT, tokenValue='8')


def test_lex_float():
    __single_token_test('4.5', TT.FLOAT, tokenValue='4.5')


def test_lex_string():
    input = 'bucket nom-8'
    expectedOutput = [
        __getTokenString('file', 1, 1, TT.STRING, 'bucket'),
        __getTokenString('file', 1, 7, TT.WHITESPACE),
        __getTokenString('file', 1, 8, TT.STRING, 'nom'),
        __getTokenString('file', 1, 11, TT.MINUS),
        __getTokenString('file', 1, 12, TT.INT, 8),
        __getTokenString('file', 1, 13, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.EOF),
    ]
    lexer = Lexer({'': ''})
    lexer.lex(input, 'file')
    output = lexer.tokenList

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_run_lexer():
    input = {
        'file': 'bucket nom-8: amount: 5 \n\t',
        'file2': 'network aaaa-#i: nombre: 2 4.5\n\t- property',
    }
    expectedOutput = [
        __getTokenString('file', 1, 1, TT.STRING, 'bucket'),
        __getTokenString('file', 1, 7, TT.WHITESPACE),
        __getTokenString('file', 1, 8, TT.STRING, 'nom'),
        __getTokenString('file', 1, 11, TT.MINUS),
        __getTokenString('file', 1, 12, TT.INT, 8),
        __getTokenString('file', 1, 13, TT.COLON),
        __getTokenString('file', 1, 14, TT.WHITESPACE),
        __getTokenString('file', 1, 15, TT.AMOUNT),
        __getTokenString('file', 1, 21, TT.COLON),
        __getTokenString('file', 1, 22, TT.WHITESPACE),
        __getTokenString('file', 1, 23, TT.INT, '5'),
        __getTokenString('file', 1, 24, TT.WHITESPACE),
        __getTokenString('file', 1, 25, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.TAB),
        __getTokenString('file', 2, 2, TT.NEWLINE),
        __getTokenString('file', 3, 1, TT.EOF),

        __getTokenString('file2', 1, 1, TT.STRING, 'network'),
        __getTokenString('file2', 1, 8, TT.WHITESPACE),
        __getTokenString('file2', 1, 9, TT.STRING, 'aaaa'),
        __getTokenString('file2', 1, 13, TT.MINUS),
        __getTokenString('file2', 1, 15, TT.VAR, 'i'),
        __getTokenString('file2', 1, 16, TT.COLON),
        __getTokenString('file2', 1, 17, TT.WHITESPACE),
        __getTokenString('file2', 1, 18, TT.STRING, 'nombre'),
        __getTokenString('file2', 1, 24, TT.COLON),
        __getTokenString('file2', 1, 25, TT.WHITESPACE),
        __getTokenString('file2', 1, 26, TT.INT, '2'),
        __getTokenString('file2', 1, 27, TT.WHITESPACE),
        __getTokenString('file2', 1, 28, TT.FLOAT, '4.5'),
        __getTokenString('file2', 1, 31, TT.NEWLINE),
        __getTokenString('file2', 2, 1, TT.TAB),
        __getTokenString('file2', 2, 2, TT.MINUS),
        __getTokenString('file2', 2, 3, TT.WHITESPACE),
        __getTokenString('file2', 2, 4, TT.STRING, 'property'),
        __getTokenString('file2', 2, 12, TT.NEWLINE),
        __getTokenString('file2', 3, 1, TT.EOF),
    ]
    lexer = Lexer(input)
    output = lexer.run()

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]


def test_run_lexer_var_in_name():
    input = {
        'file': 'bucket nom-#test \n\t toto: tata',
    }
    expectedOutput = [
        __getTokenString('file', 1, 1, TT.STRING, 'bucket'),
        __getTokenString('file', 1, 7, TT.WHITESPACE),
        __getTokenString('file', 1, 8, TT.STRING, 'nom'),
        __getTokenString('file', 1, 11, TT.MINUS),
        __getTokenString('file', 1, 13, TT.VAR, 'test'),
        __getTokenString('file', 1, 17, TT.WHITESPACE),
        __getTokenString('file', 1, 18, TT.NEWLINE),
        __getTokenString('file', 2, 1, TT.TAB),
        __getTokenString('file', 2, 2, TT.WHITESPACE),
        __getTokenString('file', 2, 3, TT.STRING, 'toto'),
        __getTokenString('file', 2, 7, TT.COLON),
        __getTokenString('file', 2, 8, TT.WHITESPACE),
        __getTokenString('file', 2, 9, TT.STRING, 'tata'),
        __getTokenString('file', 2, 13, TT.NEWLINE),
        __getTokenString('file', 3, 1, TT.EOF),
    ]
    lexer = Lexer(input)
    output = lexer.run()

    assert len(output) == len(expectedOutput)
    for i in range(len(expectedOutput)):
        assert repr(output[i]) == expectedOutput[i]



from setuptools import setup, find_packages

__version__ = '0.16.3'

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open('README.md') as rm:
    readme = rm.read()

setup(
    name='thipster',
    authors=[
        {'name': 'rcattin', 'email': 'rafa.cattin+thipster@gmail.com'},
        {'name': 'gsuquet', 'email': 'gsuquet@ippon.fr'},
    ],
    description='THipster is a tool dedicated to simplifying the difficulty associated \
with writing Terraform files. It allows users to write infrastructure as code in a \
simplified format, using either YAML (with JINJA) or the dedicated Thipster DSL.',
    long_description=readme,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Programming Language :: Python',
    ],

    download_url='https://github.com/THipster/THipster.git',
    url='https://github.com/THipster/THipster',

    version=__version__,
    install_requires=required,
    packages=find_packages(
        exclude=['pipelines'],
    ),
    extras_require={
        'test': [
            'pytest',
            'pytest-mock',
        ],
        'dev': [
            'pytest',
            'pytest-mock',
            'dagger.io',
            'pre-commit',
        ],
        'doc': [
            'sphinx',
            'myst-parser',
        ],

        'google': [
            'cdktf-cdktf-provider-google==7.0.9',
            'google-auth==2.19.1',
        ],
    },
)



from thipster.engine import THipsterException


class CDKException(THipsterException):
    pass


class CDKInvalidAttribute(CDKException):
    def __init__(self, attr: str, modelType: str, **args: object) -> None:
        super().__init__(*args)
        self.__attr = attr
        self.__modelType = modelType

    @property
    def message(self) -> str:
        return f'{self.__attr} in {self.__modelType} but not useful'


class CDKMissingAttribute(CDKException):
    @property
    def message(self) -> str:
        return 'Missing an attribute'


class CDKMissingAttributeInDependency(CDKMissingAttribute):
    pass


class CDKDependencyNotDeclared(CDKException):
    def __init__(self, depType: str, depName: str, **args: object) -> None:
        super().__init__(*args)
        self.__name = depName
        self.__type = depType

    @property
    def message(self) -> str:
        return f'{self.__type} {self.__name} not declared \
(be sure to declare it before using it)'


class CDKCyclicDependencies(CDKException):
    def __init__(self, stack: list[str], **args: object) -> None:
        super().__init__(*args)
        self.__stack = stack

    @property
    def message(self) -> str:
        return ','.join(self.__stack)
